import pygame.time
import random
## A track is a list [x_1,...,x_n], where x_i is a number. The numbers
## represent the things the track can be made of: plain rails, rails with
## people or boni, empty space, switches to the left and right.
##The following list contains the map from numbers that occur in tracks
##to their interpretations:
## Number -> Interpretation
## 0         no rails (empty space)
## 1         plain rails
## 2         switch to the track on the left
## 3         switch to the track on the right
## 4         water
## 5         hippy (a sequence of 5's combined will make one hippy)
## 6         obstacle
## 7         life bonus
## 8         score bonus

def create_new_track(length):
    track = []
    for i in range(length):
        track.append(1)
    return track

class GameState:
    def __init__(self,num_tracks,track_length):
        self.tracks = []
        self.num_tracks = num_tracks
        self.track_length = track_length
        for i in range(num_tracks):
            self.tracks.append(create_new_track(track_length))
        self.lives = 3
        ## fraction of one track element that the train passes in one
        ## millisecond:
        self.train_speed = (1.0/500.0)
        self.train_position = 0
        self.last_update = pygame.time.get_ticks()
        self.scroll_progress = 0.0
        self.to_be_added = []
        self.last_object = pygame.time.get_ticks()
        self.object_frequency = 10000
        self.switch_object_dist = 3
        self.take_switch = False
        self.train_y = 3
        self.lives = 3
        self.score = 0
        self.game_over = False
        self.paused = False
        self.paused_start = 0
        self.paused_end = 0
        self.crashed = False
        self.crash_end = 0
    def update(self):
        if self.paused == True:
            if pygame.time.get_ticks() >= self.paused_end:
                self.paused = False
                self.update()
            else:
                return
        if self.crashed == True:
            if pygame.time.get_ticks() >= self.crash_end:
                self.crashed = False
        passed_time = pygame.time.get_ticks() - self.last_update
        self.pass_time(passed_time)
        self.last_update = pygame.time.get_ticks()
        passed_time = pygame.time.get_ticks() - self.last_object
        if self.create_object(passed_time) == True:
            self.last_object = pygame.time.get_ticks()
    def toggle_switch(self):
        if self.get_switch_on_track(self.get_train_position()) != 0:
            self.take_switch = not self.take_switch
    def get_direction(self):
        if self.take_switch == False:
            return 0
        switch_dir = self.get_switch_on_track(self.get_train_position())
        return switch_dir
    def get_switch_on_track(self,track_index):
        if track_index >= 100:
            return 0
        track = self.tracks[track_index]
        for elm in track[self.train_y:]:
            if elm == 2:
                return -1
            if elm == 3:
                return 1
        return 0
    def create_object(self,milliseconds):
        if milliseconds >= self.object_frequency:
            if self.get_train_position() >= 100:
                return False
            src_track = self.get_train_position()
            if src_track == 0:
                possible_dst = [1]
            elif src_track == self.num_tracks-1:
                possible_dst = [self.num_tracks-2]
            else:
                possible_dst = [src_track-1,src_track+1]
            dst_track = random.choice(possible_dst)
            obj_function_kinds = [0,1]
            kind = random.choice(obj_function_kinds)
            obj_functions = [self.add_obstacle_or_bonus,
                             self.add_fake_obstacle_or_bonus]
            if kind == 0: ## bonus
                types = [7,8]
            else: ## obstacle
                types = [0,4,6]
            obj_func = random.choice(obj_functions)
            obj_type = random.choice(types)
            if obj_func(src_track,dst_track,obj_type) == True:
                return True
        return False
    def add_obstacle_or_bonus(self,src,dst,type):
        return self.add_object(src,dst,src,1,3,
                               self.switch_object_dist,type,1,4)
    def add_fake_obstacle_or_bonus(self,src,dst,type):
        return self.add_object(src,dst,dst,1,3,
                               self.switch_object_dist,type,1,4)
    def apply_addendum(self):
        for add in self.to_be_added:
            scroll_steps = add[2]
            ## check whether the element ought to be added right now
            if scroll_steps == 0:
                track = add[0]
                element = add[1]
                ## set the latest element in the track to the desired
                ## value
                self.tracks[track][-1] = element
                ## mark add for deletion
                add[2] = -1
            else:
                ## if not, decrease the scroll_steps count by one
                add[2] = scroll_steps - 1
        self.delete_adds()
    def delete_adds(self):
        self.to_be_added = filter((lambda (x) : (x[2] != -1)),
                                  self.to_be_added)
    def add_addendum(self,new_add):
        for add in self.to_be_added:
            if self.adds_conflict(add,new_add):
                return False
        self.to_be_added.append(new_add)
        return True
    def adds_conflict(self,add_a,add_b):
        if (add_a[0] == add_b[0]) and (add_a[2] == add_b[2]):
            return True
        return False
    def produce_addendum(self,beginning,length,track,element):
        new_adds = []
        for i in range(length):
            new_add = [track,element,beginning+i]
            new_adds.append(new_add)
        return new_adds
    def add_object(self,src_track,dst_track,obj_track,begin,
                   plain_before_switch,
                   plain_after_switch,obj_type,obj_length,plain_after_obj):
        safe_adds = self.to_be_added[0:]
        ## The future to_be_added elements for the tracks before the switch:
        before_switch_src = self.produce_addendum(begin,plain_before_switch,
                                                  src_track,1)
        before_switch_dst = self.produce_addendum(begin,plain_before_switch,
                                                  dst_track,1)
        ## The future to_be_added element for the switch:
        if dst_track == (src_track - 1):
            switch_elm = 2 ## switch to the left
        else:
            switch_elm = 3 ## switch to the right
        switch_begin = begin+plain_before_switch
        switch = self.produce_addendum(switch_begin,1,src_track,switch_elm)
        ## The future to_be_added elements for the tracks after the switch:
        after_switch_begin = switch_begin+1
        after_switch_src = self.produce_addendum(after_switch_begin,
                                                 plain_after_switch,
                                                 src_track,1)
        after_switch_dst = self.produce_addendum(after_switch_begin,
                                                 plain_after_switch,
                                                 dst_track,1)
        ## The future to_be_added elements for the object itself:
        obj_begin = after_switch_begin+plain_after_switch
        obj = self.produce_addendum(obj_begin,obj_length,obj_track,obj_type)
        ## The future to_be_added element for the tracks after the object
        ## on the destination track:
        after_obj_begin = obj_begin+obj_length
        after_obj = self.produce_addendum(after_obj_begin,
                                          plain_after_obj,
                                          dst_track,1)
        for add in (before_switch_src + before_switch_dst + \
                    switch + after_switch_src + after_switch_dst + \
                    obj + after_obj):
            if self.add_addendum(add) == False:
                self.to_be_added = safe_adds
                return False
        return True
    def pass_time(self,passed_milliseconds):
        fraction_passed = self.train_speed * passed_milliseconds
        scroll_remaining = 1.0 - self.scroll_progress
        milliseconds_remaining = scroll_remaining / self.train_speed
        if fraction_passed > scroll_remaining:
            self.remove_lowest_track_elements()
            self.add_track_elements()
            self.scroll_progress = 0.0
            self.position_train()
            self.train_events()
            self.pass_time(passed_milliseconds - milliseconds_remaining)
        else:
            self.scroll_progress = self.scroll_progress + fraction_passed
        return True
    def train_events(self):
        if self.train_position >= 10:
            return
        elm = self.tracks[self.train_position][self.train_y]
        if elm == 2:
            ## switch to the left
            if self.take_switch == True:
                self.train_position = 100 + self.train_position
        elif elm == 3:
            ## switch to the right
            if self.take_switch == True:
                self.train_position = 200 + self.train_position
        elif elm in [0,4,6]:
            self.lives = self.lives - 1
            if self.lives == -1:
                self.game_over = True
            else:
                self.pass_time((1.0)/(self.train_speed)+1.0)
                self.crashed = True
                self.pause_game(2000)
                self.crash_end = pygame.time.get_ticks() + 2000
        elif elm == 7: # life bonus:
            self.lives = self.lives + 1
        elif elm == 8: # score bonus:
            self.score = self.score + 100
    def pause_game(self,time):
        self.paused = True
        self.paused_end = pygame.time.get_ticks() + time
        self.last_update = self.last_update + time
        self.last_object = self.last_object + time
    def position_train(self):
        if (self.train_position >= 100) and (self.train_position < 200):
            new_pos = (self.train_position - 100) - 1
            self.train_position = new_pos
            self.take_switch = False
        if (self.train_position >= 200) and (self.train_position < 300):
            new_pos = (self.train_position - 200) + 1
            self.train_position = new_pos
            self.take_switch = False
    def add_track_elements(self):
        for track in self.tracks:
            track.append(1)
        self.apply_addendum()
    def remove_lowest_track_elements(self):
        for track_index in range(self.num_tracks):
            self.tracks[track_index] = self.tracks[track_index][1:]        
    def get_train_position(self):
        return self.train_position
        

## def state_matrix(state):
##     rows = []
##     for row_index in range(state.track_length):
##         maximum = state.track_length-1
##         row = []
##         for col_index in range(state.num_tracks):
##             row.append(state.tracks[col_index][maximum-row_index])
##         rows.append(row)
##     train_pos = state.get_train_position()
##     rows[train_pos][0] = 9
##     return rows

## def flip_matrix(matrix):
##     new_matrix = []
##     maximum = len(matrix) - 1
##     for row_index in range(len(matrix)):
##         new_matrix.append(matrix[maximum-row_index])
##     return new_matrix

## def print_matrix(matrix):
##     f_matrix = flip_matrix(matrix)
##     for row in f_matrix:
##         row_string = ""
##         for column in row:
##             row_string = row_string + ("%i " % column)
##         print row_string


## state = GameState(5,10)
## ## #print_matrix(state_matrix(state))
## ## #print state.scroll_progress
## ## #state.add_object(0,1,0,0,3,3,4,4,3)
## state.tracks[0][8] = 3
## ## for i in range(20):
## ##     state.pass_time(3000)
## ##     print
## ##     print state.tracks
## ##     print state.scroll_progress
## ##     print state.get_train_position()

## done = False
## last = state.tracks[0:]
## position = state.get_train_position()
## move_count = 0
## while done == False:
##     pygame.time.delay(10)
##     state.update()
##     if state.tracks != last:
##         last = state.tracks[0:]
##         move_count = move_count + 1
##         print "Moved up (%i)" % move_count
##         if move_count == 9:
##             done = True
##     if state.get_train_position() != position:
##         position = state.get_train_position()
##         print "Moved to %i" % position

